#include <stdio.h>
#include <stdlib.h> 
#include <string.h>
#include <assert.h>
#include "utils.h"
#include "fragment.h"
#include "keyword.h"
#include "dawg.h"

static fragment_cell *new_fragment_cell(const char *keyword)
{
  fragment_cell *new_frag;

  new_frag=(fragment_cell *)malloc(sizeof(fragment_cell));
  assert(new_frag != NULL);

  new_frag->keywords=new_keyword_set(keyword);
  new_frag->inf_bound=0;
  new_frag->sup_bound=0;
  new_frag->previous=NULL;
  new_frag->next=NULL;

  return new_frag;
}

static fragment append_fragment(fragment l,int inf,int sup,fragment_cell *cell)
{
  cell->next=NULL;

  if(l == NULL)
    {
      cell->previous=NULL;
      return cell;
    }

  for(;l->next != NULL;l=l->next);

  l->next=cell;
  cell->previous=l;
  l->inf_bound=inf;
  l->sup_bound=sup;

  return cell;
}

static fragment prepend_fragment(fragment_cell *cell,int inf,int sup,fragment l)
{
  cell->previous=NULL;

  if(l == NULL)
    {
      cell->next=NULL;
      return cell;
    }

  for(;l->previous != NULL;l=l->previous);

  l->previous=cell;
  cell->next=l;
  cell->inf_bound=inf;
  cell->sup_bound=sup;

  return cell;
}

void reset_fragment(fragment l,node *source)
{ 
  if(l == NULL)
    {
      return;
    }
  
  for(;l->previous != NULL;l=l->previous);

  while(l != NULL)
    {
      reset_keyword_set(l->keywords,source);
      l=l->next;
    }
}

void print_fragment(fragment l)
{
  if(l == NULL)
    {
      return;
    }

  for(;l->previous != NULL;l=l->previous);

  while(l != NULL)
    {
      fprintf(stdout,"%s",l->keywords->parent);
      //print_keyword_set(l->keywords);
      if(l->next != NULL)
	{
#ifdef SEMI_BOUNDED
	  if(l->sup_bound >= 0)
	    {
#endif
	      if((l->inf_bound != 0) || (l->sup_bound != 0))
		{
		  fprintf(stdout,"(%d,%d)",l->inf_bound,l->sup_bound);
		}
#ifdef SEMI_BOUNDED
	    }
	  else
	    {
	      fprintf(stdout,"(%d,*)",l->inf_bound);
	    }
#endif
	}
      l=l->next;
    }
}

/************************************************/
/* the string frag is assumed to be well formed */
/* i.e. it must follow the rule:                */
/* fragment = keyword | keyword(p,q)fragment    */
/************************************************/
fragment compute_fragment(const char *frag)
{
  char c;
  char *p,*q,*copy;
  int inf=0,sup=0;
  fragment l=NULL;

  copy=strdup(frag); // in order to work also with constant strings
  assert(copy != NULL);

  q=copy;
  p=q;
  while((c=*(q++)) != '\0')
    {
      switch(c)
	{
	case LEFT_PAREN:
	  *(--q)='\0';
	  l=append_fragment(l,inf,sup,new_fragment_cell(p));
	  *(q++)=LEFT_PAREN;
	  p=q;
	  while(*(q++) != COMMA);
	  *(--q)='\0';
	  inf=atoi(p);
	  *(q++)=COMMA;
	  p=q;
	  while(*(q++) != RIGHT_PAREN);
	  *(--q)='\0';
#ifdef SEMI_BOUNDED
	  if(*p == STAR)
	    {
	      sup=-1;
	    }
	  else
	    {
#endif
	      sup=atoi(p);
#ifdef SEMI_BOUNDED
	    }
#endif
	  *(q++)=RIGHT_PAREN;
	  p=q;
#ifdef SEMI_BOUNDED
	  if((sup >= 0) && (sup < inf))
#else
	  if(sup < inf)
#endif
	    {
	      int tmp;
	      tmp=inf;
	      inf=sup;
	      sup=tmp;
	    }
	  break;
	case BACKSLASH:
	  c=*(q++);
	  assert(c != '\0');
	default:
	  // nothing
	}
    }

  l=append_fragment(l,inf,sup,new_fragment_cell(p));

  free(copy);

  for(;l->previous != NULL;l=l->previous);
  return l;
}

// the real prototype of the following function is:
// int load_fragment(fragment l,int position,node *source,pattern *pat,pattern_fragment *pat_frag)
// see fragment.h for details
int load_fragment(fragment l,int position,node *source,void *pat,void *pat_frag)
{
  assert(l != NULL);
  for(;l->previous != NULL;l=l->previous);

  while(l != NULL)
    {
      if(!is_loaded(l->keywords))
	{
	  break;
	}
      l=l->next;
    }
  if(l != NULL)
    {
      if(load_keyword_set(l->keywords,position,source,l,pat,pat_frag))
	{
	  return (l->next == NULL);
	}
      else
	{
	  return (0==1);
	}
    }
  else
    {
      return (0==0);
    }
}

int previous_keyword_occurs(fragment_cell *previous_cell,int start_pos)
{
  int p,q;
  assert(previous_cell != NULL);
  p=previous_cell->inf_bound;
  q=previous_cell->sup_bound;
#ifdef SEMI_BOUNDED
  if(q < 0)
    {
      return is_element_in_bound(previous_cell->keywords->positions,q,start_pos-p);
    }
  else
    {
#endif
      return is_element_in_bound(previous_cell->keywords->positions,start_pos-q,start_pos-p);
#ifdef SEMI_BOUNDED
    }
#endif
}

void unload_fragment(fragment l)
{
  assert(l != NULL);
  for(;l->previous != NULL;l=l->previous);

  // fprintf(stderr,"unloading fragment\n");

  while(l != NULL)
    {
      unload_keyword_set(l->keywords);
      l=l->next;
    }
}

void clean_fragment(fragment l,int position)
{
  assert(l != NULL);
  for(;l->previous != NULL;l=l->previous);
  
  for(;l->next != NULL;l=l->next)
    {
      if(is_empty_int_list(l->keywords->positions))
	{
	  break;
	}
    }
  
  if(l->next != NULL) 
    {
      l=l->previous;

      for(;l != NULL;l=l->previous)
	{
	  // fprintf(stdout,"%s %d\n",l->keywords->parent,l->sup_bound);
	  // fprintf(stdout,"avant:\n");
	  // print_int_list(l->keywords->positions);
#ifdef SEMI_BOUNDED
	  if(l->sup_bound >= 0)
	    {
#endif
	      l->keywords->positions=int_list_clean(l->keywords->positions,position-l->sup_bound-l->next->keywords->length);
	      // fprintf(stdout,"aprs:\n");
	      // print_int_list(l->keywords->positions);
#ifdef SEMI_BOUNDED
	    }
#endif
	  if(!is_empty_int_list(l->keywords->positions))
	    {
		  // fprintf(stderr,"toto\n");
	      break;
	    }
	}
      // print_int_list(l->keywords->positions);
      // fprintf(stderr,"gc ");
    }
}
